#ifndef CART_TOPS
#define CART_TOPS

#include <iostream>
#include <cstdio>
#include <limits.h>
#include "mpi.h"
#include "arrays.h"

// lots of zeros for initializing dimStarts to 0
int zeros[1000];

// TODO Sub-topologies, where some coords are fixed as
// fringe or mirror, and the others are mutable.
// allows any permutation of a subset of a parent's topology
// to be used.

template <unsigned int N>
struct CartTopology {
private:
  RootArray<int, N> rankArray;
  MPI::Cartcomm cartComm;
  int thisRank, thisCoord[N];

  // create rank array, and cart communicator
  void init(const unsigned int dims[], const bool periods[]) {
    // init cart communicator
    int dims2[N];
    int n0 = MPI::COMM_WORLD.Get_size();
    int n = n0;
    for (int i = 0; i < N; i++) {
      // copy dims to int array
      dims2[i] = dims[i];
      // check are enough nodes for this topology
      n /= dims[i];
      if (n == 0) {
        n = 1;
        std::cerr << "CartTopology: " << n0 << " is not enough nodes for this toplogy:\n";
        for (int dim = 0; dim < N; dim++) { 
          n *= dims[dim];
          std::cerr << "dim " << dim << ": size " << dims[dim] << "\n";
        }
        std::cerr << "Need at least " << n << " nodes!\n";
        throw 1;
      }
    }
    //for (int i = 0; i < N; i++) printf("dim %i: size %i, period %i\n", i, dims2[i], periods[i]);
    cartComm = MPI::COMM_WORLD.Create_cart((int)N, dims2, periods, true);

    // if the communicator is non-null (i.e. this rank is in it)
    if (cartComm != MPI::COMM_NULL) {    

		  // get this process' rank and coord
		  thisRank = cartComm.Get_rank();
		  cartComm.Get_coords(thisRank, N, thisCoord);

		  // get ranks for coordinates
		  int coord[N];
		  initRanks(0, coord);

		  // check thisCoord is correct
		  if (rankArray[thisCoord] != thisRank) {
		    std::cerr << "CartTopology: thisRanks don't match: " << thisRank << " != " << rankArray[thisCoord] << "\n";
		    throw 1;
		  } 
 
    }
  }
 
  // get ranks for coordinates in topology
  void initRanks(int dim, int coord[]) {
    unsigned int dimSize = rankArray.getDimSize(dim);
    for (coord[dim] = 0; coord[dim] < dimSize; coord[dim]++) {
      if (dim == N-1) {
        // get rank for coordinate
        rankArray[coord] = cartComm.Get_cart_rank(coord);
      } else {
        // recurse...
        initRanks(dim+1, coord);
      }
    }
  }

  // display all ranks as a tree
  void printRanks(int dim, int coord[]) {
    unsigned int dimSize = rankArray.getDimSize(dim);
    for (coord[dim] = 0; coord[dim] < dimSize; coord[dim]++) {
      // display new coord at this level
      for (int i = 0; i < dim; i++) std::cout << "  ";
      if (dim == N-1) {
        // display rank for this coord
        std::cout << coord[dim] << ": ";
        std::cout << rankArray[coord] << "\n";
      } else {
        // recurse...
        std::cout << coord[dim] << "\n";
        printRanks(dim+1, coord);
      }
    }
  }

  // throws error if this not isn't in topology
  void checkInTop(const char *fname) {
    if (!isInTopology()) {
      int rank = MPI::COMM_WORLD.Get_rank();
      std::cerr << "CartTopology: calling " << fname << "; this node " << rank << " is not in topology!\n";
      throw 1;
    }
  }

  void translateRanks(int n, const int ranks1[], const MPI::Intracomm comm2, int ranks2[]) {
    // get groups
    MPI::Group group1 = cartComm.Get_group(), group2 = comm2.Get_group();
    // translate rank array
    MPI::Group::Translate_ranks(group1, n, ranks1, group2, ranks2);
  }

  int translateRank(const int rank1, const MPI::Intracomm comm2) {
    int rank2;
    translateRanks(1, &rank1, comm2, &rank2);
    //printf("on %i: rank %i to rank %i\n", thisRank, rank1, rank2);
    return rank2;
  }

public:
  // constructor, defaults to non periodic
  CartTopology(const unsigned int dims[]): rankArray(MPI::INT, dims, zeros) {
    bool periods[N];
    for (int i = 0; i < N; i++) periods[i] = false;
    init(dims, periods);
  }

  // constructor, where some axes may be periodic
  CartTopology(const unsigned int dims[], const bool periods[]): rankArray(MPI::INT, dims, zeros) {
    init(dims, periods);
  }
 
  // destructor
  ~CartTopology() {
    // free cart topology...
    //cartComm.Free();
  }

  // number of dimensions in this topology
  const unsigned int getNDims() const {
    return N;
  }

  int getRank() {
    return thisRank;
  }

  // returns true if this node is
  // in the topology
  bool isInTopology() {
    return cartComm != MPI::COMM_NULL;
  }

  // returns true if this node is on the
  // fringe of all dimensions
  bool isOnRoot() {
    for (int i = 0; i < N; i++) {
      if (thisCoord[i] != 0) return false;
    }
    return true;
  }

  // returns true if this node is on the
  // fringe in the given dimensions
  bool isOnFringe(const bool dims[]) {
    // check in toplology
    checkInTop("isOnFringe");
    // check if on fringe
    bool onFringe = true;
    for (int dim = 0; dim < N; dim++) {
      // if this is one of the fringe dims
      if (dims[dim]) {
        // check it's on it
        onFringe = (thisCoord[dim] == 0);
        if (!onFringe) return false;
      }
    }
    return true;
  }

  // returns the colour for the split comm
  // when converting the dims in changingDims
  // from fringe to mirrored. Basically groups
  // nodes into the different communicators that will
  // be used to broadcast values along.
  // i.e. selects strips orthogonal to the mirror expansion.
  // NOTE: returning a positive value does not mean
  // that this node should be involved in the mirror 
  // operation. Some colors won't be used in some broadcasts,
  // because they won't have any data in them.
  int getMirrorColor(const bool changingDims[]) {
    checkInTop("getMirrorColor");
    // for every non-changing/non-expanding dimension
    // there must be a different colour
    int color = 1; // 0 is the color to ignore!
    for (int dim = 0; dim < N; dim++) {
      if (!changingDims[dim]) {
        color += thisCoord[dim];
        color *= rankArray.getDimSize(dim);
      }
    } 
    return color;
  }

  // dim flags used to the layout of a collection
  // on a cart topology
  enum { DIM_PART_MIN=0, DIM_PART_MAX=10000, DIM_FRINGE=-1, DIM_MIRR=-2 };
  
  static const char* getDimFlagName(int flag) {
    static const char *p = "DIM_PART", *f = "DIM_FRINGE", *m = "DIM_MIRR";
    switch (flag) {
      case DIM_FRINGE: return f;
      case DIM_MIRR: return m;
      default: if (flag >= DIM_PART_MIN && flag <= DIM_PART_MAX) return p;
               else std::cerr << "getDimFlagName: should never hit default!\n"; throw 1;
    }
  }

  // returns true if there should be some local
  // data stored at this node for the given partitioning
  bool hasLocalData(const int dims[]) {
    // check this is in the toplogy
    if (!isInTopology()) return false;
    // check for dims that only store data on their fringes
    for (int dim = 0; dim < N; dim++) {
      // if data is only on the fringe in this dim
      if (dims[dim] == DIM_FRINGE) {
        // check this node is on the fringe in that dimension
        if (thisCoord[dim] != 0) return false;
      }
    }
    return true;
  }

  // mirrors the data 
  // TODO add before and after layouts
  void mirrorData(void *data, int count, MPI::Datatype type,
                  const int dimsBefore[], const int dimsAfter[]) {
    // see if we have any local data (if so we are a root)
    bool isRoot = hasLocalData(dimsBefore);
    // see if we should receive any local data (if so we are root or a receiver)
    bool isReceiver = hasLocalData(dimsAfter);
    //printf("rank %i: isReceiver=%i, isRoot=%i\n", thisRank, isReceiver, isRoot);
    // if is root or receiver do broadcast
    if (isRoot || isReceiver) {
      // get changing dims
      bool expandingDims[N];

      // get the root coord
      // the changing dims should all be 0 (at fringe)
      // the others should be taken from this node's coord.
      int rootCoord[N];
  
      for (int dim = 0; dim < N; dim++) {
        // if dim going fringe -> mirror, then is expanding
        if (dimsBefore[dim] == DIM_FRINGE &&
            dimsAfter[dim] == DIM_MIRR) {
          expandingDims[dim] = true;
          rootCoord[dim] = 0;
          continue;
        } else {
          rootCoord[dim] = thisCoord[dim];
          expandingDims[dim] = false;
        }
        // check we're not trying to change a dim in any
        // other way (e.g. can't go from mirr -> fringe, or part -> mirr etc)
        if (dimsBefore[dim] != dimsAfter[dim]) {
          std::cerr << "CartTopology: mirrorData: can only convert dims from fringe to mirror!\n"
                  << "Can't convert dim " << dim << " from " << getDimFlagName(dimsBefore[dim])
                  << " to " << getDimFlagName(dimsAfter[dim]) << "\n";
          throw 1;
        }
      }
      // get color
      int color = getMirrorColor(expandingDims);
      // make communicator
      MPI::Intracomm mirrComm = cartComm.Split(color, thisRank);
      // get root rank in cartComm
      int cartRootRank = rankArray[rootCoord];
      //printf("rank %i: coord={%i, %i}, rootRank=%i, rootCoord={%i, %i}, color=%i\n", thisRank, thisCoord[0], thisCoord[1], cartRootRank, rootCoord[0], rootCoord[1], color);
      // check if this is root, that the rootRank is thisRank
      if (isRoot && cartRootRank != thisRank) {
        std::cerr << "CartTopology: mirrorData: rootRank " << cartRootRank << " doesn't match thisRank " << thisRank << "!\n";
        throw 1;
      }
      // translate rank for mirror comm
      int rootRank = translateRank(cartRootRank, mirrComm);
      // do broadcast...
      mirrComm.Bcast(data, count, type, rootRank);
      // TODO free communicator?
    } else {
      // for all nodes that won't receive any data:
      cartComm.Split(0, thisRank);
    }
  } 

  // mirror array
  template <class V, unsigned int NArr>
  void mirrorArray(
    SubArray<V, NArr> *arrIn,
    SubArray<V, NArr> *arrOut,
    const int dimsBefore[], const int dimsAfter[]) 
  {
    mirrorArray<V, NArr>(arrIn, arrOut, dimsBefore, dimsAfter, NULL);
  }

  // mirror array
  template <class V, unsigned int NArr>
  void mirrorArray(
    SubArray<V, NArr> *arrIn,
    SubArray<V, NArr> *arrOut,
    const int dimsBefore[], const int dimsAfter[], V* outBuf) 
  {
    // check this either sends or receives data
    if (!hasLocalData(dimsBefore) && !hasLocalData(dimsAfter)) return;

    // mirror array bounds
    int boundsLen=arrIn->boundsPackedSize();
    int boundsBuf[boundsLen];
    arrIn->packBounds(boundsBuf);
    mirrorData(boundsBuf, boundsLen, MPI::INT, dimsBefore, dimsAfter);
    int arrStarts[NArr];
    unsigned int arrLens[NArr];
    arrIn->unpackStarts(boundsBuf, arrStarts);
    arrIn->unpackSizes(boundsBuf, arrLens);

    // if sending from here, assign arrIn to arrOut
    if (hasLocalData(dimsBefore)) {
      *arrOut = *arrIn;
      std::cout << "mirrorArr: data send from " << thisRank << "\n";
    }
    // if we are receiving here, then prepare result array
    else {
      // if already init, check size is big enough
      if (arrOut->getParent() != NULL && arrOut->getParent()->isInit()) {
        #ifdef DEBUG
        printf("mirrorArr: initing outArr provided to be big enough.\n");
        #endif
        arrOut->initArr(arrOut->getParent(), arrLens, arrStarts); // initArr checks bounds are subset
      } 
      // otherwise init new root array to store result
      else {
        arrOut->initArr(arrIn->getElementType(), arrLens, arrStarts);
        #ifdef DEBUG
        printf("mirrorArr: initing a new root array to receive into.\n");
        #endif
      }
      std::cout << "mirrorArr: recv data at " << thisRank << "\n";
    }

    // assign underlying buffer to out buff
    if (outBuf != NULL) {
      arrOut->parent->elems = outBuf;
    }

    // broadcast data
    V* dataBuf = &(*arrOut)[arrOut->getDimStarts()];  
    mirrorData(dataBuf, 1, arrOut->getType(), dimsBefore, dimsAfter);
  }

  // scatter vectors (already partitioned by a stream consumer)

  // scatter array (slice up and distribute)
  // NArr - number of dimensions in the array
  // NPart - number of dimensions changing from fringe to partitioned
  template <class V, unsigned int NArr, unsigned int NPart>
  void scatterArrayBlocked(
    SubArray<V, NArr> *arrIn,
    SubArray<V, NArr> *arrOut,
    const int dimsBefore[], const int dimsAfter[]) {
    // see if we should be sending data
    bool isRoot = hasLocalData(dimsBefore);
    // see if we should receive any local data (if so we are root or a receiver)
    bool isReceiver = hasLocalData(dimsAfter);

    #ifdef DEBUG
    printf("%i: scatter array isRoot=%i, isReceiver=%i\n", thisRank,  isRoot, isReceiver);
    for (int i = 0; i < NArr; i++) {
      printf("%i: dimsBefore[%i] = %i, dimsAfter[%i] = %i\n", 
         thisRank, i, dimsBefore[i], i, dimsAfter[i]);
    }
    for (int i = 0; i < N; i++) {
      printf("%i: thisCoord[%i] = %i\n", thisRank, i, thisCoord[i]);
    }
    #endif 

    // if this node is not involved, return 
    if (!isRoot && !isReceiver) return;

    // get the root coord
    // the changing dims should all be 0 (at fringe)
    // the others should be taken from this node's coord.
    int rootCoord[N];

    // find which dims are changing from fringe to partition(arrayDim)
    int numDimsToPart = 0, arrDimsToPart[NPart], dimToPartAlong[NPart];
    for (int i = 0; i < N; i++) {
      // if layout is changing, check how
      if (dimsBefore[i] != dimsAfter[i]) {
        if (dimsBefore[i] == DIM_FRINGE && 
            dimsAfter[i] >= DIM_PART_MIN &&
            dimsAfter[i] <= DIM_PART_MAX) {
          // add a dim to partition to list
          arrDimsToPart[numDimsToPart] = dimsAfter[i];
          dimToPartAlong[numDimsToPart] = i;
          numDimsToPart++;
          // set this dim to 0 in rootCoord
          rootCoord[i] = 0;
        } else {
          // shouldn't be any other kind of layout change
          std::cerr << "CartTopology::scatterArrayBlocked: can only convert dims from fringe to partitioned|"
                    << "Can't convert dim " << i << " from " << getDimFlagName(dimsBefore[i]) << " to "
                    << getDimFlagName(dimsAfter[i]) << "\n";
          throw 1;
        }
      } else {
        rootCoord[i] = thisCoord[i];
      }
    }

    // check numDimsToPart matches constant parameter NPart
    if (numDimsToPart != NPart) {
      std::cerr << "CartTopology::scatterArrayBlocked: numDims to partition " 
                << numDimsToPart << " doesn't match template parameter " << NPart << "!\n";
      throw 1;
    }

    #ifdef DEBUG
    for (int i = 0; i < N; i++) {
      printf("%i: rootCoord[%i] = %i, thisCoord[%i] = %i\n", thisRank, i, rootCoord[i], i, thisCoord[i]);
    }
    #endif 

    // if is a sender
    if (isRoot) { 
      // display parts to dim
      printf("scattering array:\n");
      for (int i = 0; i < numDimsToPart; i++) {
        printf("part arr dim %i along %i\n", arrDimsToPart[i], dimToPartAlong[i]);
      }
      printf("\n");
  
      // get num partitions in each dim
      unsigned int numPartsPerDim[N];
      for (int i = 0; i < numDimsToPart; i++) {
        numPartsPerDim[i] = rankArray.getDimSize(dimToPartAlong[i]); //arrDimsToPart[i]);
      }

      // TODO - DEBUG, PARTITIONING WRONG!

      // partition array 
      int partsStarts[NPart];
      for (int i = 0; i < NPart; i++) partsStarts[i] = 0;
      RootArray<SubArray<V, NArr>, NPart> *parts = new RootArray<SubArray<V, NArr>, NPart>(MPI::DATATYPE_NULL, numPartsPerDim, partsStarts);
      //arrIn->template partition<NPart>(arrDimsToPart, numPartsPerDim, &parts);
      partitionArray<V, NArr, NPart>(arrDimsToPart, numPartsPerDim, arrIn, parts);

      // display partitions
      for (typename RootArray<SubArray<V, NArr>, NPart>::idx_iterator it = parts->beginIndex();
           it != parts->endIndex(); it++) {
        printf("part ");
        printCoord<NPart>(it.getIndex());
        printf(" ");
        it.getElement()->printDims(); 
        printf("\n");
      } 

      // start sending partitions to receiver nodes
      int partCoord[NPart], topCoord[N];
      int numRequests = 1;
      for (int i = 0; i < NPart; i++) numRequests *= parts->getDimSize(i);
      numRequests *= 2; // each send must send array bounds, then data
      MPI::Request requests[numRequests];
      startSends<V, NArr, NPart>(0, partCoord, topCoord, dimToPartAlong, parts, arrOut, requests);

      #ifdef DEBUG  
      printf("scatter started %i sends...\n", numRequests);
      #endif 
      
      // finish sending
      MPI::Request::Waitall(numRequests, requests);

      #ifdef DEBUG  
      printf("scatter finished %i sends...\n", numRequests);
      #endif 

      // deallocate parts
      delete parts;
    }

    // if is a receiver 
    else if (isReceiver && !isRoot) {
      // get sender rank
      int senderRank = rankArray[rootCoord];

      #ifdef DEBUG
      printf("%i: scatter receiving array bounds...\n", thisRank);
      #endif

      // receive array bounds
      int arrLenStart[NArr*2];
      cartComm.Recv(arrLenStart, NArr*2, MPI::INT, senderRank, 0);
      unsigned int arrLen[NArr];
      for (int i = 0; i < NArr; i++) {
        arrLen[i] = arrLenStart[i+NArr];
      }

      #ifdef DEBUG
      printf("%i: scatter received array bounds.\n", senderRank);
      #endif

      printf("receiving part");
      printf(" from ");
      printCoord<N>(rootCoord);
      printf(" rank %i ", senderRank); 
      printf(" to ");
      printCoord<N>(thisCoord);
      printf(" rank %i ", thisRank);
      for (int i = 0; i < NArr; i++) printf("; start %i, len %i", arrLenStart[i], arrLen[i]);
      printf("\n");

      // if no underlying root array, then create one and receive to that
      if (arrOut->getParent() == NULL) {
        // create a root array, and subarray
        RootArray<V, NArr> *rootArr = new RootArray<V, NArr>(arrIn->getElementType(), arrLen, &arrLenStart[0]);
        arrOut->initArr(rootArr, arrLen, &arrLenStart[0]);
        // TODO WHEN TO DEALLOCATE ROOT ARR?
        // REFCOUNTING?

        // receive into them
        V* buf = &(*rootArr)[rootArr->getDimStarts()];  
        cartComm.Recv(buf, 1, rootArr->getType(), senderRank, 1);
        #ifdef DEBUG
        printf("%i: scatterArrBlocked: allocated a root array on the heap\n", thisRank);
        #endif
      }
      // receive into the root array given 
      else { 

        // if refers to root array that is not initialized
        if (!arrOut->getParent()->isInit()) {
          // has a root array, but that array hasn't allocated 
          // an elems array yet. So init it, so it does allocate space to
          // receive to
          RootArray<V, NArr> *rootArr = arrOut->getParent();
          rootArr->initArr(arrLen, &arrLenStart[0]);
          arrOut->initArr(rootArr);
          #ifdef DEBUG
          printf("%i: scatterArrBlocked: initialized root array provided, to receive into\n", thisRank);
          #endif
        }
 
        // receive into the array provided. 
        arrOut->initArr(arrOut->getParent(), arrLen, &arrLenStart[0]);
        V* buf = &(*arrOut)[arrOut->getDimStarts()];
        cartComm.Recv(buf, 1, arrOut->getType(), senderRank, 1);

        #ifdef DEBUG
        printf("%i: scatterArrBlocked: received into root array\n", thisRank);
        #endif
      }
    }
  }

  /*template <class V, unsigned int NArr, unsigned int NPart>
  void scatterArrayBlocked(
    RootArray<V, NArr> *arrIn,
    RootArray<V, NArr> *arrOut,
    const int dimsBefore[], const int dimsAfter[]) 
  {
    // initialize subarray for arrIn
    SubArray<V, NArr> sarrIn(arrIn);
    SubArray<V, NArr> sarrOut(arrOut);
    
    // call scatter
    this->template scatterArrayBlocked<V, NArr, NPart>(&sarrIn, &sarrOut, dimsBefore, dimsAfter);
  }*/

protected:
  // sends an array of arrays, to their corresponding nodes
  template <class V, unsigned int NArr, unsigned int NPart>
  void startSends(int dim, int partCoord[], int topCoord[], const int dimsToPartAlong[], 
    RootArray<SubArray<V, NArr>, NPart> *parts, 
    SubArray<V, NArr> *arrOut,
    MPI::Request requests[]) {
    // init topCoord
    if (dim == 0) {
      for (int i = 0; i < N; i++) topCoord[i] = thisCoord[i];
      for (int i = 0; i < NPart; i++) partCoord[i] = 0;
    }

    // recursively iterate over partitions 
    if (dim == NPart) {
      // base case: send
      // get subarray & it's type
      SubArray<V, NArr> *part = &(*parts)[partCoord];

      // rank of node to send to
      int destRank = this->rankArray[topCoord];

      // idx of request
      int reqIdx = partCoord[0];
      for (int i = 1; i < NPart; i++) {
        reqIdx *= parts->getDimSize(i);
        reqIdx += partCoord[i];
      } 

      // start async send
      printf("send part");
      printCoord<NPart>(partCoord);
      printf(" from ");
      printCoord<N>(thisCoord);
      printf(" rank %i ", thisRank); 
      printf(" to ");
      printCoord<N>(topCoord);
      printf(" rank %i ", destRank);  
      part->printDims(); 
      printf("\n");

      // if sending to self, then assign result immediately
      if (thisRank == destRank) {
        *arrOut = *part;
      }
      // send to slave 
      else {
        #ifdef DEBUG
        printf("%i: sending partition array bounds to %i\n", thisRank, destRank);
        #endif
 
        // send starts and lengths
        int startLen[NArr*2];
        for (int i = 0; i < NArr; i++) startLen[i] = part->getDimStart(i);
        for (int i = 0; i < NArr; i++) {startLen[i+NArr] = part->getDimSize(i);
        printf("; start: %i, size %i", startLen[i], startLen[i+NArr]);}
        requests[reqIdx] = cartComm.Isend(startLen, NArr*2, MPI::INT, destRank, 0);

        // send array data
        MPI::Datatype partTy = part->getType();
        V* buf = &(*part)[part->getDimStarts()];
        requests[reqIdx*2] = cartComm.Isend(buf, 1, partTy, destRank, 1);
      }

        printf("\n");
      
    } else {
      // topology dim to partition this array dim over
      int topDim = dimsToPartAlong[dim];

      // recursive case, iterate along this dimension
      for (int i = 0; i < parts->getDimSize(dim); i++) {
        topCoord[topDim] = i;
        partCoord[dim] = i;
        startSends<V, NArr, NPart>(dim+1, partCoord, topCoord, dimsToPartAlong, parts, arrOut, requests);
      }
    }
  }

public:
  // RedT reducer class is what's used to do local reductions
  template <class RedT, class V1, class V2, unsigned int N1, unsigned int N2>
  void reduceArray(
    RedT reducer, const MPI::Datatype& outElemTy,
    SubArray<V1, N1> *arrIn, RootArray<V2, N2> *arrOut,
    const int dimsBefore[], const int dimsAfter[]) 
  {
    // see if we should be sending data or receiving ot both
    bool isSender = hasLocalData(dimsBefore);
    bool isRoot = hasLocalData(dimsAfter);

    // if neither then do nothing
    if (!isSender && !isRoot) return;

    // check root is also a sender
    if (isRoot && !isSender) {
      std::cerr << "CartTopology::reduceArray: receiver " << thisRank << " must also be a sender!\n";
      throw 1;
    }

    // get changing dims
    bool reducingDims[N];
    int numDimsToReduce = 0;

    // get the root coord
    // the changing dims should all be 0 (at fringe)
    // the others should be taken from this node's coord.
    int rootCoord[N];
  
    for (int dim = 0; dim < N; dim++) {
      // if dim going part -> fringe, then is reducing
      if (dimsBefore[dim] >= DIM_PART_MIN &&
          dimsBefore[dim] <= DIM_PART_MAX &&
          dimsAfter[dim] == DIM_FRINGE) {
        reducingDims[dim] = true;
        rootCoord[dim] = 0;
        numDimsToReduce++;
        continue;
      } else {
        rootCoord[dim] = thisCoord[dim];
        reducingDims[dim] = false;
      }
      // check we're not trying to change a dim in any
      // other way
      if (dimsBefore[dim] != dimsAfter[dim]) {
        std::cerr << "CartTopology::reduceArray: can only convert dims from part to fringe!\n"
                << "Can't convert dim " << dim << " from " << getDimFlagName(dimsBefore[dim])
                << " to " << getDimFlagName(dimsAfter[dim]) << "\n";
        throw 1;
      }
    }

    // check num dims to reduce is valid
    if (numDimsToReduce != N1 - N2) {
      std::cerr << "CartTopology::reduceArray: num dims to reduce "
                << numDimsToReduce << " doesn't match N1 " << N1
                << " - N2" << N2 << " = " << (N1 - N2) << "!\n";
      throw 1;
    }

    // make communicator for this strip
    int color = getMirrorColor(reducingDims);
    
    // make communicator
    MPI::Intracomm redComm = cartComm.Split(color, thisRank);
    // get root rank in cartComm
    int cartRootRank = rankArray[rootCoord];
    // check if this is root, that the rootRank is thisRank
    if (isRoot && cartRootRank != thisRank) {
      std::cerr << "CartTopology::reduceArray: rootRank " << cartRootRank << " doesn't match thisRank " << thisRank << "!\n";
      throw 1;
    }
    // translate rank for mirror comm
    int rootRank = translateRank(cartRootRank, redComm);

    printf("root ");
    printCoord<N>(rootCoord);
    printf(" rank %i ;", rankArray[rootCoord]);
    printCoord<N>(thisCoord);
    printf(" , rank %i is color %i\n", thisRank, color);

    // reduce local part
    RootArray<V2, N2> temp(outElemTy);
    reducer(arrIn, &temp);

    printf("temp on %i\n", thisRank);
    for (typename RootArray<V2, N2>::idx_iterator it = temp.beginIndex();
         it != temp.endIndex(); it++) {
      printf("temp on %i: ", thisRank);
      printCoord<N2>(it.getIndex());
      printf(" val %i\n", *it.getElement());
    }

    // if at root, init local result
    V2* recvBuf = NULL;
    if (isRoot) {
      // get starts and lens for local result
      unsigned int *lens = temp.getDimSizes();
      int *starts = temp.getDimStarts();

      // check if out array is init
      if (!arrOut->isInit()) {
        // init it
        arrOut->initArr(lens, starts);
      }

      // init subarray
      recvBuf = &(*arrOut)[starts];
    }

    // do inter-node reduce
    V2* buf = &temp[temp.getDimStarts()];
    redComm.Reduce(buf, recvBuf, 1, temp.getType(), *reducer.getOp(), rootRank);

 }

 template <class V, unsigned int NArr>
  void allGatherArray(
    SubArray<V, NArr> *arrIn,
    SubArray<V, NArr> *arrOut,
    const int dimsBefore[], const int dimsAfter[]) 
  {
    // put dim ids into lists for those
    // going from part to mirr, and those
    // staying part.
    int numChanging=0, numPart=0;
    int changingDims[N], partDims[N];
    bool changingDims1[N];
    for (int d = 0; d < N; d++) {
      changingDims1[d] = false;
      if (dimsBefore[d] >= DIM_PART_MIN &&
          dimsBefore[d] <= DIM_PART_MAX &&
          dimsAfter[d] == DIM_MIRR) {
        changingDims[numChanging] = d;
        changingDims1[d] = true;
        numChanging++;
      } else if (dimsBefore[d] != dimsAfter[d]) {
        std::cerr << "CartTopology::allGatherArray: dim " << d << " can't change from "
                  << dimsBefore[d] << " to " << dimsAfter[d] << "\n";
        throw 1;
      }
      if (dimsAfter[d] >= DIM_PART_MIN &&
          dimsAfter[d] <= DIM_PART_MAX) {
        partDims[numPart] = d;
        numPart++;
      }
    }

    // check at least 1 dim is changing 
    if (numChanging == 0) {
        std::cerr << "CartTopology::allGatherArray: no dims changing from part to mirr!\n";
        throw 1;
    }

    // see if we have any local data 
    bool isSender = hasLocalData(dimsBefore);
    // see if we should receive any local data (if so we are root or a receiver)
    bool isReceiver = hasLocalData(dimsAfter);
    // if sending/receiving some data
    if (isSender || isReceiver) {
      // get color
      int color = getMirrorColor(changingDims1);
      // make communicator
      MPI::Intracomm comm = cartComm.Split(color, thisRank);
      
      // allgather array bounds
      int boundsLen=arrIn->boundsPackedSize();
      int boundsBuf[boundsLen];
      arrIn->packBounds(boundsBuf);
      int partCount = comm.Get_size();
      int recvBounds[boundsLen*partCount];
      comm.Allgather(boundsBuf, boundsLen, MPI::INT, recvBounds, boundsLen, MPI::INT);

      // unpack bounds
      int arrStarts[NArr*partCount];
      unsigned int arrLens[NArr*partCount];
      for (int p = 0; p < partCount; p++) {
        arrIn->unpackStarts(&recvBounds[p*boundsLen], &arrStarts[p*NArr]);
        arrIn->unpackSizes(&recvBounds[p*boundsLen], &arrLens[p*NArr]);
        //printf("unpack on %i part %i start %i len %i\n", 
        //   thisRank, p, arrStarts[p*NArr], arrLens[p*NArr]);
      }

      // find max and min of all 
      int arrMaxs[NArr], arrMins[NArr];
      unsigned int outArrLens[NArr];
      for (int d = 0; d < NArr; d++) {
        arrMaxs[d] = INT_MIN;
        arrMins[d] = INT_MAX; 
        for (int p = 0; p < partCount; p++) {
          int i = (p*NArr)+d;
          if ((int)arrLens[i] <= 0) continue; // check for empty dim 
          if (arrStarts[i] < arrMins[d]) arrMins[d] = arrStarts[i];
          int arrEnd = arrStarts[i] + arrLens[i] - 1;
          if (arrEnd > arrMaxs[d]) arrMaxs[d] = arrEnd;
          #ifdef DEBUG
          printf("sizes recvd on %i: part %i dim %i start %i end %i\n",
            thisRank, p, d, arrStarts[i], arrEnd);
          #endif
        }
        outArrLens[d] = arrMaxs[d] - arrMins[d] + 1;
        #ifdef DEBUG
        printf("max/min: %i [%i..%i] len %i\n", d, arrMins[d], arrMaxs[d], outArrLens[d]);
        #endif
      } 

      // allocate output array
      arrOut->initArr(arrIn->getElementType(), outArrLens, arrMins);

      // find non-empty partitions
      int numNonEmpty = 0;
      int nonEmptyParts[partCount];
      for (int p = 0; p < partCount; p++) {
        bool isEmpty = false;
        for (int d = 0; d < NArr; d++) {
          #ifdef DEBUG
          printf("arrlen of part %i dim %i is %i\n", p, d, (int)arrLens[(p*NArr)+d]);
          #endif
          if ((int)arrLens[(p*NArr)+d] <= 0) {
            isEmpty = true; break;
          }
        }
        if (!isEmpty) {
          nonEmptyParts[numNonEmpty] = p;
          numNonEmpty++;
        }
      }

      // allocate receive views
      SubArray<V, NArr> *views = new SubArray<V, NArr>[numNonEmpty];
      for (int i = 0; i < numNonEmpty; i++) {
        int p = nonEmptyParts[i];
        views[i].initArr(arrOut, &arrLens[p*NArr], &arrStarts[p*NArr]);
        #ifdef DEBUG
        printf("on %i: view %i: dim0: [%i..%i] len %i\n", thisRank, i, views[i].getDimStart(0), views[i].getDimEnd(0), views[i].getDimSize(0));
        printf("on %i: view %i: dim1: [%i..%i] len %i\n", thisRank, i, views[i].getDimStart(1), views[i].getDimEnd(1), views[i].getDimSize(1));
        #endif
      }

      // check if this is empty
      bool thisEmpty = false; 
      for (int d = 0; d < NArr; d++) {
        if ((int)arrIn->getDimSize(d) <= 0) {
          #ifdef DEBUG
          printf("on %i dim %i is empty\n", thisRank, d);
          #endif
          thisEmpty = true;
          break;
        }
      }

      // if this is non-empty, send it to all other procs
      if (!thisEmpty) {
        MPI::Request sendReqs[partCount];
        for (int p = 0; p < partCount; p++) {
          sendReqs[p] = comm.Isend(arrIn->getStartPointer(), 
            1, arrIn->getType(), p, 0);
          #ifdef DEBUG
          printf("sending from %i to %i\n", thisRank, p);
          #endif
        }
      }

      // receive from all non-empty procs   
      MPI::Request recvReqs[numNonEmpty];
      for (int i = 0; i < numNonEmpty; i++) {
        int p = nonEmptyParts[i];
        #ifdef DEBUG
        printf("receiving from %i to %i into:\n", p, thisRank);
        printf("dim0: [%i..%i] len %i\n", views[i].getDimStart(0), views[i].getDimEnd(0), views[i].getDimSize(0));
        printf("dim1: [%i..%i] len %i\n", views[i].getDimStart(1), views[i].getDimEnd(1), views[i].getDimSize(1));
        #endif
        recvReqs[i] = comm.Irecv(views[i].getStartPointer(),
          1, views[i].getType(), p, 0);
      }
    
      // wait for all receives to complete
      MPI::Request::Waitall(numNonEmpty, recvReqs);

      // free views
      delete [] views;

    } else {
      // for all nodes that won't receive any data:
      cartComm.Split(0, thisRank);
    }
    
  }

  // display on std::out
  void print() {
    std::cout << "CartTopology<" << N << ">\n";
    for (int dim = 0; dim < N; dim++) {
      std::cout << "dim " << dim << ": " << rankArray.getDimSize(dim) << "\n";
    }
    int coord[N];
    printRanks(0, coord);
  }

};
#endif
